//
// nettop 
//
// Copyright (c) 2000
//           Scott R Parish <sRp@portico.org>, OR  97119-9201.
//       All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer as
//    the first lines of this file unmodified.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY Scott R Parish ``AS IS'' AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
// OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL Scott R Parish BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
// NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

//
// TODO
// 
// * optimize
// * figure ports out better
// * add paging (for viewing all packets)
// * show udp ports
// * show more about 802.3 
// * have a mode where you only see statistics from the last x time period
// * allow for configuring which traffic is viewed
// * use autoconf/automake 
// * allow for collapsing and expanding of sub-types (like tcp/udp/etc)
// * move eth types/ip types/etc out to seperate file (user defineable)
// * get tcp/udp from /etc/services
// * find why screen gets a line messed up every so often
// * remove segfaults on screen resize
// * allow user to specify a filter on the command line
//

/*
 * Rewritten for non-threaded DOS/djgpp.
 * Reformatted using GNU Indent.
 * Based on thr 14 Feb 2000 version
 *
 * G. Vanem 2000 <giva@bgnett.no>
 */

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <limits.h>
#include <time.h>
#include <slang.h>
#include <pcap.h>
#include <tcp.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>

#include "node.h"

#ifdef __GNUC__
extern void SLsmg_printf (char *, ...) __attribute__ ((format (printf, 1, 2)));
#endif

#define REFRESH_TIME   500 /* 0.5sec */

#define VAL_FORMAT     "%3Lu.%2.2Lu%% %8s"   /* maybe use "llu" ? */
#define avg_print(x,y) if (y) \
                              SLsmg_printf(" %6Lu", (uint64)(x/y)); \
                         else SLsmg_printf(" %6s", "??")

#define xprintf(s,x)     x = strlen(s); SLsmg_printf ("%s", s)


/* Colors
 */
#define ethernet_c 1
#define ip_c       2
#define tcp_c      3
#define bar_c      4
#define udp_c      5
#define total_c    6
#define header_c   7

pcap_t *pcap_start (void);

void ipprint (int *y);
void tcpprint (int *y);
void alarmed (void);
void dieing (int);
void ui_run (void);
void ts_run (void);
void sprintf_nice (char *str, uint64 data);
void change_update_wait (void);
void show_help (void);
void change_sort_type (void);
void pause_display (void);
void clear_count (void);
void usage (void);
void show_stat (void);

struct node ether = { NULL, NULL, -1, 0ULL, 0ULL };
struct node ip    = { NULL, NULL, -1, 0ULL, 0ULL };
struct node tcp   = { NULL, NULL, -1, 0ULL, 0ULL };

char  *device       = NULL;
char   show_banner  = 52;
int    start_time   = 0;
int    quit         = 0;
int    update_wait  = 1000000;  /* how long to wait between screen updates */
u_char sort_type    = 'p';      /* either (p)acket or (s)ize */

pcap_t *pd;

int main (int argc, char **argv)
{
  int ch;

  while ((ch = getopt (argc, argv, "i:")) != -1)
  {
    switch (ch)
    {
      case 'i':
           device = optarg;
           break;
      case '?':
      default:
           usage();
           exit (1);
    }
  }
  pd = pcap_start();
  signal (SIGINT, dieing);
  while (!quit)
  {
    ui_run();
    ts_run();
  }
  return (0);
}

void ts_run (void)
{
  struct pcap_pkthdr hdr;
  struct node       *t;
  struct ip          iphdr;
  struct tcphdr      tcphdr;
  u_char *ptr;
  u_short eth_type;
  int     ip_head_len;
  int     port;

  if ((ptr = pcap_next (pd, &hdr)) == NULL)
     return;

  eth_type = ntohs (*(u_short*)&ptr[12]);

  if (eth_type < 0x05dd)
  {                           // this is an 802.3 packet
    u_short type;

    memcpy (&type, &ptr[18], 2);
    type = ntohs (type);
    t = get_node (&ether, type);
    if (!t)
       t = new_node (&ether, type);
    t->count++;
    t->size += hdr.len;
  }
  else                        // this is a ethernet packet
  {
    t = get_node (&ether, eth_type);
    if (!t)
       t = new_node (&ether, eth_type);
    t->count++;
    t->size += hdr.len;

    switch (eth_type)
    {
      case 0x0800:
           memcpy (&iphdr, ptr + 14, sizeof(iphdr));
           ip_head_len = iphdr.ip_hl * 4;
           t = get_node (&ip, iphdr.ip_p);
           if (!t)
              t = new_node (&ip, iphdr.ip_p);
           t->count++;
           t->size += hdr.len;

           if (iphdr.ip_p == 6)
           {
             memcpy (&tcphdr, ptr + 14 + ip_head_len, sizeof(tcphdr));
             port = (ntohs (tcphdr.th_sport) < ntohs (tcphdr.th_dport) ?
                     ntohs (tcphdr.th_sport) : ntohs (tcphdr.th_dport));

             t = get_node (&tcp, port);
             if (!t)
                t = new_node (&tcp, port);
             t->count++;
             t->size += hdr.len;
           }
           break;

      default:
    }
  }
}

/*
 * this function, is called from ui_run(), and is
 * used to update the screen
 */
void screen_update (void)
{
  struct node *sorted[24] = { 0 };
  uint64 sump, sums, valp, vals;
  int    i, time_diff, a, b, c;
  int    x = 0, y = 1, maxx, maxy;
  char   str[40];
  time_t t;

  sump = count_sump (&ether);
  sums = count_sums (&ether);
  largestn (&ether, sorted, 24, sort_type);
  SLsmg_gotorc (0, 0);

  x = 0;

  if (show_banner)
  {
    SLsmg_set_color (show_banner);
    show_banner = show_banner > 50 ? show_banner - 1 : show_banner;
    xprintf ("Nettop: by sRp@linux.com - c 2000 BSD", x);
  }
  else                    // print clock
  {
    SLsmg_set_color (header_c);
    t = time (0);
    SLsmg_printf ("%s", ctime (&t));
    x = 24;
  }

  //print counter
  for (; x < SLtt_Screen_Cols - 10; x++)
     SLsmg_printf (" ");

  time_diff = time (0) - start_time;
  a = time_diff / 3600;
  b = time_diff % 3600 / 60;
  c = (time_diff % 3600) % 60;

  SLsmg_set_color (header_c);
  SLsmg_printf ("%4d:%2.2d:%2.2d", a, b, c);
  x = 0;

  SLsmg_gotorc (y++, x);

  x = 11;
  maxy = SLtt_Screen_Rows;
  maxx = SLtt_Screen_Cols;

  //print bar
  SLsmg_set_color (bar_c);
  SLsmg_printf ("  %%pkts    total   %%size    total sz/pkt");

  SLsmg_printf ("%*s", 29, "type/port");
  for (i = (x += 4); i < maxx; i++)
    SLsmg_printf (" ");
  SLsmg_gotorc (y++, 0);

  SLsmg_set_color (total_c);
  sprintf_nice (str, sump);
  SLsmg_printf ("%3d.%2.2d%% %8s", 100, 0, str);
  sprintf_nice (str, sums);
  SLsmg_printf (" %3d.%2.2d%% %8s", 100, 0, str);
  if (sump > 0)
     SLsmg_printf (" %6u", (u_int)(sums / sump));
  SLsmg_gotorc (y - 1, maxx - 20);
  SLsmg_printf ("total");
  SLsmg_gotorc (y++, 0);

  /* print percentages
   */
  for (i = 0; i < 24; i++)
  {
    if (sorted[i])
    {
      /* the first node in our array is there as a place holder
       * we don't want to print it out.
       */
      if (sorted[i]->type == 0xffffffff)
         continue;

      if (y > maxy - 1)
      {
        SLsmg_refresh();
        return;
      }

      SLsmg_set_color (ethernet_c);
      valp = sorted[i]->count * 10000ULL / sump;
      sprintf_nice (str, sorted[i]->count);

      SLsmg_printf (VAL_FORMAT, valp / 100ULL, valp % 100ULL, str);
      vals = sorted[i]->size * 10000ULL / sums;
      sprintf_nice (str, sorted[i]->size);
      SLsmg_printf (" " VAL_FORMAT, vals / 100ULL, vals % 100ULL, str);
      avg_print (sorted[i]->size, sorted[i]->count);
      SLsmg_gotorc (y - 1, maxx - 20);

      switch (sorted[i]->type)
      {
        case 0x0800:
             SLsmg_printf ("ipv4         ");
             ipprint (&y);
             break;
        case 0x0806:
             SLsmg_printf ("arp          ");
             break;
        case 0x6002:
             SLsmg_printf ("dec mop      ");
             break;
        case 0x6003:
             SLsmg_printf ("dec dna-route");
             break;
        case 0x6004:
             SLsmg_printf ("dec lat      ");
             break;
        case 0x8035:
             SLsmg_printf ("rarp         ");
             break;
        case 0x809b:
             SLsmg_printf ("appletalk DDP");
             break;
        case 0x8137:
             SLsmg_printf ("novell ipx-nw");
             break;
        case 0x86dd:
             SLsmg_printf ("ipv6         ");
             break;
        default:
             SLsmg_printf ("0x%4.4x       ", sorted[i]->type);
      }
      SLsmg_gotorc (y++, 0);
    }
  }
  SLsmg_refresh();
}

void ipprint (int *cury)
{
  struct node *sorted[24] = { 0 };
  int    y = *cury, maxx, maxy, j;
  char   str[40];
  uint64 valp, vals, sump, sums;

  sump = count_sump (&ether);
  sums = count_sums (&ether);

  largestn (&ip, sorted, 24, sort_type);

  maxy = SLtt_Screen_Rows;
  maxx = SLtt_Screen_Cols;

  for (j = 0; j < 24; j++)
  {
    if (sorted[j])
    {
      if (sorted[j]->type == -1)
        continue;

      SLsmg_set_color (ip_c);

      if (y > maxy - 1)
      {
        *cury = y;
        return;
      }

      SLsmg_gotorc (y, 0);

      valp = sorted[j]->count * 10000ULL / sump;
      sprintf_nice (str, sorted[j]->count);
      SLsmg_printf (VAL_FORMAT, valp / 100ULL, valp % 100ULL, str);
      vals = sorted[j]->size * 10000ULL / sums;
      sprintf_nice (str, sorted[j]->size);
      SLsmg_printf (" " VAL_FORMAT, vals / 100ULL, vals % 100ULL, str);
      avg_print (sorted[j]->size, sorted[j]->count);
      SLsmg_gotorc (y++, maxx - 20);

      switch (sorted[j]->type)
      {
        case 00:
             SLsmg_printf ("reserved     ");
             break;
        case 01:
             SLsmg_printf ("icmp         ");
             break;
        case 02:
             SLsmg_printf ("igmp         ");
             break;
        case 06:
             SLsmg_printf ("tcp          ");
             tcpprint (&y);
             break;
        case 17:
             SLsmg_printf ("udp          ");
             break;
        case 21:
             SLsmg_printf ("prm          ");
             break;
        case 31:
             SLsmg_printf ("MFE NSP      ");
             break;
        case 34:
             SLsmg_printf ("3pc          ");
             break;
        case 70:
             SLsmg_printf ("gopher       ");
             break;
        case 113:
             SLsmg_printf ("auth         ");
             break;
        default:
             SLsmg_printf ("%d           ", sorted[j]->type);
      }
    }
  }
  *cury = y;
}

void tcpprint (int *cury)
{
  struct node *sorted[24] = { 0 };
  int    y = *cury, maxx, maxy, j;
  char   str[40];
  uint64 vals, valp, sums, sump;

  sump = count_sump (&ether);
  sums = count_sums (&ether);

  largestn (&tcp, sorted, 24, sort_type);

  maxy = SLtt_Screen_Rows;
  maxx = SLtt_Screen_Cols;

  SLsmg_set_color (tcp_c);
  for (j = 0; j < 24; j++)
  {
    if (sorted[j])
    {
      if (sorted[j]->type == -1)
        continue;

      if (y > maxy - 1)
      {
        *cury = y;
        return;
      }

      SLsmg_gotorc (y, 0);

      valp = sorted[j]->count * 10000ULL / sump;
      sprintf_nice (str, sorted[j]->count);
      SLsmg_printf (VAL_FORMAT, valp / 100ULL, valp % 100ULL, str);
      vals = sorted[j]->size * 10000ULL / sums;
      sprintf_nice (str, sorted[j]->size);
      SLsmg_printf (" " VAL_FORMAT, vals / 100ULL, vals % 100ULL, str);
      avg_print (sorted[j]->size, sorted[j]->count);
      SLsmg_gotorc (y++, maxx - 20);

      switch (sorted[j]->type)
      {
        case 20:
             SLsmg_printf ("ftp-data     ");
             break;
        case 21:
             SLsmg_printf ("ftp          ");
             break;
        case 22:
             SLsmg_printf ("ssh          ");
             break;
        case 25:
             SLsmg_printf ("smtp         ");
             break;
        case 79:
             SLsmg_printf ("finger       ");
             break;
        case 80:
             SLsmg_printf ("http         ");
             break;
        case 110:
             SLsmg_printf ("pop3         ");
             break;
        case 139:
             SLsmg_printf ("netbios-ssn  ");
             break;
        case 220:
             SLsmg_printf ("imap3        ");
             break;
        case 515:
             SLsmg_printf ("printer      ");
             break;
        default:
             SLsmg_printf ("%d           ", sorted[j]->type);
      }
    }
  }
  *cury = y;
}

void dieing (int sig)
{
#ifndef DONT_RESTORE_SCREEN
  SLsmg_reset_smg();
  SLang_reset_tty();
#endif
  if (sig == 0)
    show_stat();
  pcap_close (pd);
  exit (0);
}

void ui_run (void)
{
  static DWORD timeout = 0L;

  /* do this once per REFRESH_TIME
   */
  if (timeout && !chk_timeout(timeout))
     return;

  timeout = set_timeout (REFRESH_TIME);

  if (start_time == 0)   /* first time called */
  {
    start_time = time (0);

    SLtt_get_terminfo();
    SLang_init_tty (-1, 0, 0);
    SLsmg_init_smg();

    SLtt_set_color (0,         "normal","lightgray", "default");
    SLtt_set_color (ethernet_c,"level1","brightmagenta", "default");
    SLtt_set_color (ip_c,      "level2","brightgreen", "default");
    SLtt_set_color (tcp_c,     "level3","brightcyan", "default");
    SLtt_set_color (bar_c,     "bar",   "blue", "lightgray");
    SLtt_set_color (udp_c,     "udp",   "brightcyan", "default");
    SLtt_set_color (total_c,   "total", "yellow", "default");
    SLtt_set_color (header_c,  "total", "white", "default");
    SLtt_set_color (50,        "total", "gray", "default");
    SLtt_set_color (51,        "total", "lightgray", "default");
    SLtt_set_color (52,        "total", "white", "default");
  }

  if (SLang_input_pending(0))
  {
    switch (SLang_getkey())
    {
      case 's':
           change_update_wait();
           break;
      case 'c':
           clear_count();
           break;
      case 'x':
      case 'q':
           dieing (0);
           quit = 1;
           break;
      case '?':
      case 'h':
           show_help();
           break;
      case 'o':
           change_sort_type();
           break;
      case ' ':
           pause_display();
           break;
      case '\x0C':
           SLsmg_cls();
           break;
    }
  }
  if (show_banner && time(NULL) - start_time > 3)
     show_banner = 0;
  screen_update();
}

pcap_t *pcap_start (void)
{
  struct bpf_program fcode;
  int    datalink, snaplen = 200;
  char   errbuf[PCAP_ERRBUF_SIZE];
  pcap_t *pd;
  bpf_u_int32 localnet, netmask;

  if (!device)
  {
    device = pcap_lookupdev (errbuf);
    if (!device)
    {
      fprintf (stderr, "Can not open device: %s\n", errbuf);
      dieing (1);
    }
  }

  pd = pcap_open_live (device, snaplen, 1, 500, errbuf);
  if (!pd)
  {
    fprintf (stderr, "Can not open device: %s\n", errbuf);
    dieing (1);
  }

  if (pcap_lookupnet (device, &localnet, &netmask, errbuf) < 0)
  {
    fprintf (stderr, "pcap_lookupnet: %s\n", errbuf);
    dieing (1);
  }

  if (pcap_compile (pd, &fcode, " ", 0, netmask) < 0)
  {
    fprintf (stderr, "pcap_compile: %s\n", pcap_geterr (pd));
    dieing (1);
  }

  if (pcap_setfilter (pd, &fcode) < 0)
  {
    fprintf (stderr, "pcap_setfilter: %s\n", pcap_geterr (pd));
    dieing (1);
  }

  datalink = pcap_datalink (pd);
  if (datalink < 0)
  {
    fprintf (stderr, "pcap_datalink: %s\n", pcap_geterr (pd));
    dieing (1);
  }
  return (pd);
}

void sprintf_nice (char *str, uint64 data)
{
  if (data / 1024ULL == 0)
    sprintf (str, "%8u", (u_int)data);

  else if (data / 1024ULL < 1024ULL)
    sprintf (str, "%7.2fk", data / 1024.0);

  else if (data / 1048576ULL < 1048576ULL)
    sprintf (str, "%7.2fm", data / 1048576.0);

  else
    sprintf (str, "%7.2fg", data / 1073741824.0);
}

void change_update_wait (void)
{
  char str[40];
  int   x, i = 0;

  SLsmg_set_color (0);
  SLsmg_gotorc (0, 0);
  SLsmg_printf ("Please enter a new delay time in seconds:");
  for (x = 41; x < SLtt_Screen_Cols; x++)
     SLsmg_printf (" ");
  SLsmg_refresh();

  while (i < 10 && SLang_input_pending (25))
  {
    x = SLang_getkey();
    switch (x)
    {
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
      case '0':
      case '.':
           str[i++] = x;
           SLsmg_gotorc (0, 41 + i);
           SLsmg_write_char (x);
           SLsmg_refresh();
           break;
      case 127:             //backspace
           SLsmg_gotorc (0, 41 + i);
           SLsmg_write_char (' ');
           SLsmg_refresh();
           str[--i] = 0;
      case '\r':
           str[i] = 0;
           update_wait = (1000000 * atof (str));
           break;
    }
  }
}

void change_sort_type (void)
{
  int x, i = 0;

  SLsmg_set_color (0);
  SLsmg_gotorc (0, 0);
  SLsmg_printf ("Change sorting to [p]ackets or [s]ize:");

  for (x = 41; x < SLtt_Screen_Cols; x++)
      SLsmg_printf (" ");

  SLsmg_refresh();

  while (i < 10 && SLang_input_pending (25))
  {
    x = SLang_getkey();
    switch (x)
    {
      case 'p':
           sort_type = 'p';
           return;
      case 's':
           sort_type = 's';
           return;
    }
  }
}

void show_help (void)
{
  int y = 0;

  SLsmg_cls();
  SLsmg_gotorc (y++, 0);

  SLsmg_set_color (header_c);

  SLsmg_printf ("Nettop: written by Scott Parish");
  SLsmg_gotorc (y++, 0);
  SLsmg_printf ("Copyright 2000, BSD");
  SLsmg_gotorc (y++, 0);
  SLsmg_printf ("Compiled %s, %s", __TIME__, __DATE__);
  y += 2;
  SLsmg_gotorc (y++, 0);

  SLsmg_printf ("  Key     Effect");
  SLsmg_set_color (0);
  SLsmg_gotorc (y++, 0);
  SLsmg_printf ("   s      change update delay time (default 1 sec)");
  SLsmg_gotorc (y++, 0);
  SLsmg_printf ("   o      change the order of sort (by [p]ackets or [s]ize)");
  SLsmg_gotorc (y++, 0);
  SLsmg_printf ("   c      clear the data, and start all counting over)");
  SLsmg_gotorc (y++, 0);
  SLsmg_printf ("q or x    exit nettop");
  SLsmg_gotorc (y++, 0);
  SLsmg_printf ("? or h    this screen");
  SLsmg_gotorc (y++, 0);
  SLsmg_printf ("<space>   pause display from updating (but keep sniffing)");
  SLsmg_gotorc (y++, 0);

  SLsmg_gotorc (y++, 0);
  SLsmg_printf (" %%pkts    what fraction of total packets are from this type");
  SLsmg_gotorc (y++, 0);
  SLsmg_printf (" total    total number of packets from this type");
  SLsmg_gotorc (y++, 0);
  SLsmg_printf (" %%size    what fraction by size of all packets are type");
  SLsmg_gotorc (y++, 0);
  SLsmg_printf (" sz/pkt   What is the average size of each packet");
  SLsmg_gotorc (y++, 0);

  SLsmg_gotorc (SLtt_Screen_Rows - 2, 0);
  SLsmg_printf ("Press any key to return to program");
  SLsmg_refresh();

  SLang_getkey();
  SLsmg_cls();
}

void pause_display (void)
{
  int x, center;

  SLsmg_gotorc (0, 0);
  SLsmg_set_color (header_c);
  center = SLtt_Screen_Cols / 2 - 8;
  for (x = 0; x < center; x++)
    SLsmg_printf (" ");

  SLsmg_printf ("--== paused ==--");
  for (; x < SLtt_Screen_Cols; x++)
    SLsmg_printf (" ");

  SLsmg_refresh();
  SLang_getkey();
}

void clear_count (void)
{
  ndelete (&ether);
  ndelete (&ip);
  ndelete (&tcp);
  start_time = time (0);
  SLsmg_cls();
}

void show_stat (void)
{
  struct pcap_stat stat;

  if (pcap_stats (pd, &stat) < 0)
     fprintf (stderr, "pcap_stats: %s\n", pcap_geterr (pd));
  else
  {
    fprintf (stderr, "packets received by filter: %-6lu\n", stat.ps_recv);
    if (stat.ps_recv)
    {
      fprintf (stderr, "packets dropped by kernel:  %-6lu (%.1f%%)\n",
               stat.ps_drop, 100.0 * (double) stat.ps_drop / (double) stat.ps_recv);
      fprintf (stderr, "packets dropped by driver:  %-6lu (%.1f%%)\n",
               stat.ps_ifdrop, 100.0 * (double) stat.ps_ifdrop / (double) stat.ps_recv);
    }
  }
}

/*
 * usage - print the command line args options
 */
void usage (void)
{
  fprintf (stderr, "Usage: nettop [-i dev]\n");
}

